/* 
 * Please read the COPYRIGHT notice included at the top level
 */
/*
 * augment.c,v 1.2 1996/09/10 03:12:19 arsharma Exp
 * Purpose : To write out augmentation into the output file
 */

#include <stdio.h>
#include <stddef.h>
#include <assert.h>

#include "doctor_t.h"
#include "lex.h"
#include "memref.h"
#include "instr.h"
#include "mnemonics_def.h"
#include "event.h"
#include "internals.h"
#include "augment.h"
#include "registers.h"

extern bool_t default_data_movement, count_references;
extern char *aug_buffer;
extern int aug_end;

/* This becomes necessary in the stack_offset cases */
static char operand[MAX_OP_SIZE];

/* called only if instr has a memory reference */
void
augment_instruction (instr_t *instr)
{
    unsigned mnemonic_code;

    mnemonic_code = instr->mnemonic->representation;

    /* figure out if this is a floating point instruction */
    instr->floating_point = (mnemonic_code >= F2XM1)
	&& (mnemonic_code <= FYL2XP1);

    check_references (instr);
}

/*
 * Writes out the augmentation that is appropriate for the
 * instruction.
 * Please refer to Algo.data_movement for a detailed description of
 * the algorithm used for the data movement
 */
void
check_references (instr_t *instr)
{
    ref_t *ref1, *ref2;
    bool_t data_movement;

    ref1 = instr->mem_ref[0];
    ref2 = instr->mem_ref[1];

    data_movement = data_movement_safe (instr);

    /* case 1 there's only one reference */
    if ((ref1) && not (ref2)) {
	if (ref1->read) {

	    augment ("/	A read\n");

	    if (count_references) {
		augment ("\tinc	read_ref\n");
		augment ("\tinc	count_ref\n");
	    }
	    else if (data_movement) {
		augment_push_regs ();
		augment_put_stuff (instr, ref1);
		augment_pop_regs ();
		augment_ctxsw_intel (instr);
		augment_push_regs ();
		augment_move_data (instr, ref1);
		augment_pop_regs ();
		augment_done (instr);
		/* The original instruction goes here */
		augment_push_regs ();
		augment_restore_memory (instr);
		augment_pop_regs ();
	    }
	    else {
		/* normal case */
		augment_push_regs ();
		augment_put_stuff (instr, ref1);
		augment_pop_regs ();
		augment_ctxsw_intel (instr);
	    }
	}
	else {
	    /* if ref1->write */
	    augment ("/	A write\n");

	    if (count_references) {
		augment ("\tinc	write_ref\n");
		augment ("\tinc	count_ref\n");
	    }
	    else if (data_movement) {
		augment_push_regs ();
		/* save the memory location */
		augment_move_data (instr, ref1);
		augment_pop_regs ();
		augment_done (instr);
		/* The original instruction goes here */
		augment_push_regs ();
		augment_put_stuff (instr, ref1);
		augment_put_more_stuff (instr, ref1);
		augment_restore_memory (instr);
		augment_pop_regs ();
		augment_ctxsw_intel (instr);
	    }
	    else {
		/* normal case */
		augment_push_regs ();
		augment_put_stuff (instr, ref1);
		augment_pop_regs ();
		augment_ctxsw_intel (instr);
	    }
	}
    }
    else if (ref1 && ref2) {
	/* case 2 there are two references, both reads */
	if ((ref1->read) && (ref2->read)) {

	    augment ("/	two reads\n");

	    if (count_references) {
		augment ("\tinc	read_ref\n");
		augment ("\tinc	read_Ref\n");
		augment ("\tinc	count_ref\n");
		augment ("\tinc	count_ref\n");
	    }
	    else if (data_movement) {
		augment_push_regs ();
		augment_put_stuff (instr, ref1);
		augment_pop_regs ();
		augment_ctxsw_intel (instr);
		augment_push_regs ();
		augment_move_data (instr, ref1);
		/* no popping here */
		augment_put_stuff (instr, ref2);
		augment_pop_regs ();
		augment_ctxsw_intel (instr);
		augment_push_regs ();
		augment_move_data (instr, ref2);
		augment_pop_regs ();
		augment_done (instr);
		/* The original instruction goes here */
		augment_push_regs ();
		augment_restore_memory (instr);
		augment_pop_regs ();
	    }
	    else {
		/* normal case */
		augment_push_regs ();
		augment_put_stuff (instr, ref1);
		augment_pop_regs ();
		augment_ctxsw_intel (instr);
		augment_push_regs ();
		augment_put_stuff (instr, ref2);
		augment_pop_regs ();
		augment_ctxsw_intel (instr);
	    }
	}
	else if ((ref1->read) && (ref2->write)) {
	    /* case 3 there are two references, one read and one write */
	    augment ("/	A read and a write\n");

	    if (count_references) {
		augment ("\tinc	read_ref\n");
		augment ("\tinc	write_ref\n");
		augment ("\tinc	count_ref\n");
		augment ("\tinc	count_ref\n");
	    }
	    else if (data_movement) {
		augment_push_regs ();
		augment_put_stuff (instr, ref1);
		augment_pop_regs ();
		augment_ctxsw_intel (instr);
		augment_push_regs ();
		augment_move_data (instr, ref1);
		augment_pop_regs ();
		augment_push_regs ();
		augment_move_data (instr, ref2);
		augment_pop_regs ();
		augment_done (instr);
		/* The original instruction goes here */
		augment_push_regs ();
		augment_put_stuff (instr, ref2);
		augment_put_more_stuff (instr, ref2);
		/* restore everything - we need to recompute the vaddr */
		augment_pop_regs ();
		augment_push_regs ();
		augment_restore_memory (instr);
		augment_pop_regs ();
		augment_ctxsw_intel (instr);
	    }
	    else {
		/* normal case */
		augment_push_regs ();
		augment_put_stuff (instr, ref1);
		augment_pop_regs ();
		augment_ctxsw_intel (instr);
		augment_push_regs ();
		augment_put_stuff (instr, ref2);
		augment_pop_regs ();
		augment_ctxsw_intel (instr);
	    }
	}
    }
    else
	/* we have a new case here */
	assert (0);
}

void
augment_push_regs ()
{
    /* let's find an efficient way later */
    augment ("\tpushfl\n");
    augment ("\tpushl %%eax\n");
    augment ("\tpushl %%ebx\n");
}

char *unused_reg;

/*
 * puts       data into pcur_thread->pevent structure
 * Specifically, updates the type,size,vaddr fields
 */
void
augment_put_stuff (instr_t *instr, ref_t *ref)
{
    /*
     * move the       virtual address into %eax. This is done without
     * disturbing any of the registers, to ensure that the right virtual
     * address is loaded into %eax
     */

    /* get the correct stack address */
    if (ref->stack) {
	if (ref->read) {
	    sprintf (operand, ref->format_string, ref->stack_offset);
	    ref->source = operand;
	}
	else if (ref->write) {
	    /* It's a write */
	    sprintf (operand, ref->format_string, ref->stack_offset);
	    ref->destination = operand;
	}
    }

    augment ("\tleal ");
    if (ref->read)
	augment ("%s", ref->source);
    else
	augment ("%s", ref->destination);
    augment (",%%eax\n");

    /* get the offset of the `pevent' field into %ebx */
    GET_PEVENT (ebx);

    /* set the event type (r/w) */
    if (ref->read)
	augment ("\tmovl $%d,%d(,%%ebx,1)\n",
		 instr->lock ? E_LOCK_READ : E_READ | E_READ_OFFSET,
		 EVENT_TYPE_OFFSET);
    else if (ref->write)
	augment ("\tmovl $%d,%d(,%%ebx,1)\n",
		 instr->lock ? E_LOCK_WRITE : E_WRITE |
		 E_WRITE_OFFSET,
		 EVENT_TYPE_OFFSET);

    /* move the operand size into the `size' field */
    augment ("\tmovl $%d,%d(,%%ebx,1)\n",
	     instr->mnemonic->operand_size, EVENT_SIZE_OFFSET);

    /* move virtual address into the `vaddr' field */
    augment ("\tmovl %%eax,%d(,%%ebx,1)\n", EVENT_VADDR_OFFSET);

    /* Update the thread clock */
    augment_safe_update_clock ();
}

void
augment_pop_regs ()
{
    augment ("\tpopl %%ebx\n");
    augment ("\tpopl %%eax\n");
    augment ("\tpopfl \n");
}

/* writes out code to do the context switch */
void
augment_ctxsw_intel (instr_t *instr)
{
    augment ("\tpushl $pcur_thread\n");
    augment ("\tpushl psim_thread\n");
    augment ("\tpushl pcur_thread\n");
    augment ("\tcall ctxsw_intel\n");
    augment ("\tadd  $12,%%esp\n");
}

/* For writes, we pass the `value' too to the backend */
void
augment_put_more_stuff (instr_t *instr, ref_t *ref)
{
    if (ref->write) {
	/* move the value into pevent.value */
	augment ("\tpushl (%%eax)\n");
	augment ("\tpopl %d(,%%ebx,1)\n", EVENT_VALUE_OFFSET);
	if (instr->double_data) {
	    augment ("\tpushl 4(%%eax)\n");
	    augment ("\tpopl %d(%%ebx)\n", EVENT_VALUE_OFFSET + 4);
	}
    }
}

/*
 * Saves the memory locations involved and moves the data to/from the
 * backend
 */
void
augment_move_data (instr_t *instr, ref_t *ref)
{
    if (ref->write) {

	/* get the stack virtual address correctly */
	if (ref->stack) {
	    sprintf (operand, ref->format_string, ref->stack_offset);
	    ref->destination = operand;
	}

	/* get the vaddr in %eax */
	augment ("\tleal %s,%%eax\n", ref->destination);

	/* if ref is the first reference */
	if (instr->mem_ref[0] == ref) {
	    /* save the memory location in tmp1 */
	    augment ("\tpushl %%ebx\n");
	    augment ("\tpushl (%%eax)\n");
	    GET_PTHREAD (ebx);
	    augment ("\tpopl %d(%%ebx)\n", THREAD_TMP1_OFFSET);
	    if (instr->double_data) {
		augment ("\tpushl 4(%%eax)\n");
		augment ("\tpopl %d(%%ebx)\n", THREAD_TMP1_OFFSET + 4);
	    }
	    augment ("\tpopl %%ebx\n");
	}
	else {
	    /* if it is the second, use tmp2 */
	    augment ("\tpushl %%ebx\n");
	    augment ("\tpushl (%%eax)\n");
	    GET_PTHREAD (ebx);
	    augment ("\tpopl %d(%%ebx)\n", THREAD_TMP2_OFFSET);
	    if (instr->double_data) {
		augment ("\tpushl 4(%%eax)\n");
		augment ("\tpopl %d(%%ebx)\n", THREAD_TMP2_OFFSET + 4);
	    }
	    augment ("\tpopl %%ebx\n");
	}
    }
    else if (ref->read) {

	GET_PEVENT (ebx);

	/* move the vaddr into eax */
	augment ("\tmovl %d(%%ebx),%%eax\n", EVENT_VADDR_OFFSET);

	if (instr->mem_ref[0] == ref) {
	    /* save the memory location in tmp1 */
	    augment ("\tpushl %%ebx\n");
	    augment ("\tpushl (%%eax)\n");
	    GET_PTHREAD (ebx);
	    augment ("\tpopl %d(%%ebx)\n", THREAD_TMP1_OFFSET);
	    if (instr->double_data) {
		augment ("\tpushl 4(%%eax)\n");
		augment ("\tpopl %d(%%ebx)\n", THREAD_TMP1_OFFSET + 4);
	    }
	    augment ("\tpopl %%ebx\n");
	}
	else {
	    /* use tmp2 */
	    augment ("\tpushl %%ebx\n");
	    augment ("\tpushl (%%eax)\n");
	    GET_PTHREAD (ebx);
	    augment ("\tpopl %d(%%ebx)\n", THREAD_TMP2_OFFSET);
	    if (instr->double_data) {
		augment ("\tpushl 4(%%eax)\n");
		augment ("\tpopl %d(%%ebx)\n", THREAD_TMP2_OFFSET + 4);
	    }
	    augment ("\tpopl %%ebx\n");
	}

	/* move pevent.value into ebx and then into (mem) */
	switch (instr->mnemonic->operand_size) {
	    case 1:
		augment ("\tmovb %d(%%ebx),%%bl\n", EVENT_VALUE_OFFSET);
		augment ("\tmovb %%bl,(%%eax)\n");
		break;
	    case 2:
		augment ("\tmovw %d(%%ebx),%%bx\n", EVENT_VALUE_OFFSET);
		augment ("\tmovw %%bx,(%%eax)\n");
		break;
	    case 4:
		augment ("\tmovl %d(%%ebx),%%ebx\n", EVENT_VALUE_OFFSET);
		augment ("\tmovl %%ebx,(%%eax)\n");
		break;
	    case 8:
		augment ("\tpushl %%ebx\n");
		augment ("\tmovl %d(%%ebx),%%ebx\n", EVENT_VALUE_OFFSET);
		augment ("\tmovl %%ebx,(%%eax)\n");
		augment ("\tpopl %%ebx\n");
		augment ("\tmovl %d(%%ebx),%%ebx\n", EVENT_VALUE_OFFSET + 4);
		augment ("\tmovl %%ebx,4(%%eax)\n");
		break;
	    default:
		assert (0);
		break;
	}
    }
}

/* restores the saved memory locations, after the data movement */
void
augment_restore_memory (instr_t *instr)
{
    ref_t *ref1, *ref2;

    ref1 = instr->mem_ref[0];
    ref2 = instr->mem_ref[1];

    if ((ref1->read) || (ref1->write)) {

	if (ref2) {
	    /* get the correct stack offset */
	    if (ref1->stack) {
		if (ref1->read) {
		    sprintf (operand, ref1->format_string,
			     ref1->stack_offset);
		    ref1->source = operand;
		}
		else if (ref1->write) {
		    /* It's a write */
		    sprintf (operand, ref1->format_string,
			     ref1->stack_offset);
		    ref1->destination = operand;
		}
	    }

	    /* get the  `vaddr' */
	    augment ("\tleal ");
	    if (ref1->read)
		augment ("%s", ref1->source);
	    else
		augment ("%s", ref1->destination);
	    augment (",%%eax\n");
	}
	else if (ref1->read) {
	    /*
	     * In the case of a       write, we already have ebx and eax setup
	     * properly, so this part is unnecessary
	     */

	    /* get pevent in ebx */
	    GET_PEVENT (ebx);

	    /* move the vaddr into eax */
	    augment ("\tmovl %d(%%ebx),%%eax\n", EVENT_VADDR_OFFSET);
	}
	/* restore the memory location from tmp1 */
	augment ("\tpushl %%ebx\n");
	GET_PTHREAD (ebx);
	augment ("\tpushl %d(%%ebx)\n", THREAD_TMP1_OFFSET);
	augment ("\tpopl (%%eax)\n");
	if (instr->double_data) {
	    augment ("\tpushl %d(%%ebx)\n", THREAD_TMP1_OFFSET + 4);
	    augment ("\tpopl 4(%%eax)\n");
	}
	augment ("\tpopl %%ebx\n");
    }

    if (ref2 && ((ref2->read) || (ref2->write))) {

	/* get pevent in ebx */
	GET_PEVENT (ebx);

	/* move the vaddr into eax */
	augment ("\tmovl %d(%%ebx),%%eax\n", EVENT_VADDR_OFFSET);

	/* restore the memory location from tmp2 */
	augment ("\tpushl %%ebx\n");
	GET_PTHREAD (ebx);
	augment ("\tpushl %d(%%ebx)\n", THREAD_TMP2_OFFSET);
	augment ("\tpopl (%%eax)\n");
	if (instr->double_data) {
	    augment ("\tpushl %d(%%ebx)\n", THREAD_TMP2_OFFSET + 4);
	    augment ("\tpopl 4(%%eax)\n");
	}
	augment ("\tpopl %%ebx\n");
    }

}

/*
 * Inserts augmentation       for updating the thread time. The counting is
 * done according to the `base' time specified in the pentium manual.
 *
 * Warning: Assumes the ebx has pcur_thread, when it is called.
 */

void
augment_update_clock ()
{
    if (static_execution_time) {
	augment ("\tmovl $%d,static_execution_time\n", static_execution_time);
	augment ("\tfldl %d(%%ebx)\n", THREAD_TIME_OFFSET);
	augment ("\tfiaddl static_execution_time\n");
	augment ("\tfstpl %d(%%ebx)\n", THREAD_TIME_OFFSET);

	/* reset the time to zero */
	static_execution_time = 0;
    }
}

/*
 * Inserts augmentation for updating the instruction count
 */
void
augment_update_instr_count()
{
    extern int instr_count;

    if (instr_count) {
	augment("\tpushfl\n");
        augment ("\taddl $%d,doctor_instr_count\n",instr_count);
	augment("\tpopfl\n");
        instr_count = 0;
    }
}

/*
 * This       one is the same as augment_update_clock, but doesn't assume
 * anything about ebx.
 */

void
augment_safe_update_clock ()
{
    if (static_execution_time) {
	augment_push_regs ();
	GET_PTHREAD (ebx);
	augment ("\tmovl $%d,static_execution_time\n", static_execution_time);
	augment ("\tfldl %d(%%ebx)\n", THREAD_TIME_OFFSET);
	augment ("\tfiaddl static_execution_time\n");
	augment ("\tfstpl %d(%%ebx)\n", THREAD_TIME_OFFSET);
	augment_pop_regs ();

	/* reset the time to zero */
	static_execution_time = 0;
    }
}

/* checks if it is safe to do the data movement */
bool_t
data_movement_safe (instr_t *instr)
{
    /* indirect jumps are not handled */
    if ((instr->mnemonic->representation == JMP)
	&& (instr->modes[0] == INDIRECT))
	return FALSE;

    /* sizes greater than 4 bytes are not handled */
    if ((instr->mnemonic->operand_size > 8))
	return FALSE;

    return default_data_movement;
}
