/* 
 * Please read the COPYRIGHT notice included at the top level
 */
/*
 * lex.l,v 1.1.1.1 1996/07/18 22:39:57 arsharma Exp
 *
 * This	is the input to	the lexical analyzer
 * While the characters	are being read,	they are also stuffed into the
 * 'output buffer' and flushed at the end of basic blocks
 */
%{

#include "doctor_t.h"
#include "keyword_t.h"
#include  "mnemonics_def.h"
#include "lex.h"
#include "symtab.h"
#include "memref.h"
#include "instr.h"
#include "write.h"

#include <stdio.h>
#include <ctype.h>
#include <varargs.h>

extern unsigned start_first, end_first, start_second, end_second;

/*
 * This       indicates that a mnemonic has been seen and everything else,
 * including mnemonics on the current line should be recognized as IDs
 * rather than mnemonics. Had to do this, because x86 unix compilers
 * do not prefix an underscore to the variable names.
 */

bool_t seen_mnemonic = FALSE, seen_second = FALSE;

/*
 * Another trick to make sure that commas appearing inside the
 *  brackets don't deceive the lexer into assuming that the first
 *  operand has actually ended
 */
bool_t fake_comma = FALSE;

%}

White_space	[ \t]
Hex_digit	[0-9a-fA-F]
Digit		[0-9]
Letter		[a-zA-Z_]
Letter_Or_Digit	[a-zA-Z_0-9]
Label_letter	[a-zA-Z_0-9.]
Label_quirks	[^\n,]
Asterisk	[*]
Slash		[/]
Other		.
Non_colon	[^:\n]
String		"[^\n"]*"

%%


[ \t]		{
    /* " to fool the emacs string parser */
	/* Do nothing white space */
	token_store();;
}

[\n]		{

	token_store();
	seen_mnemonic = FALSE;
	seen_second = FALSE;
	end_second = 0;
	return NEWLINE;
}

"," 		{
	/*
	 *  A comma indicates the end of the
	 *  first operand and the beginning
	 *  of	the second
	 */

	if (!fake_comma)
		end_first	= save_pos();
	token_store();
	if ((!fake_comma) && (!seen_second)) {
		seen_second = TRUE; 
		start_second = save_pos();
	}
	else if (!fake_comma) {
		end_second = save_pos() - 2;
	}
	return COMMA;
}

{Letter}({Letter}|{Digit})*    	{
	token_store();
	return screen();
}

("+"|"-")?{Digit}{Digit}*		{
	token_store();
	return CONSTANT;
}

"$"[^,\n]*		       {
	token_store();
	return VALUE;
}

^{White_space}*"."[^L]{Other}* {
	/*
	 * every line, that	begins with a
	 * DOT and anything	other than a
	 * `L' is a	directive
	 */
	token_store();
	return DIRECTIVE;
}

^{White_space}*"call aug_on_\n"	{
	return AUG_ON;
}

^{White_space}*"call aug_off_\n" {
	return AUG_OFF;
}

"."			       {
	token_store();
	return DOT;
}

{Asterisk}		       {
	token_store();
	return ASTERISK;
}

":"			       {
	token_store();
	return COLON;
}

"("			       {
	fake_comma = TRUE;
	token_store();
	return OB;
}

")"			       {
	fake_comma = FALSE;
	token_store();
	return CB;
}

"$"			       {
	token_store();
	return DOLLAR;
}

"/"[^\n]*		       {
	token_store();
	return COMMENT;
}

((.L{Label_quirks}*)|({Label_letter}+))":" {
	token_store();
	return LABELDEF;
}

"%"{Letter}({Letter}|{Digit})*("("{Digit}")")? {
	/*	checking for registers */
	token_store();
	return screen();
}

%%

/* look  in lex.h for an explanation of the various fields */
static mnemonic_table_t mnemonic_table[] =
{
#include "mnemonic_table.h"
};

char last_id[30];
extern unsigned start_pos, start_instr;
mnemonic_table_t *last_mnem;
/*
 * This       routine checks if the yytext is a keyword. If so, it returns
 * the enum const corresponding to the keyword. Otherwise it checks if
 * it is a mnemonic, if it is,  it returns MNEMONIC. Otherwise it
 * returns ID
 */

static int
screen ()
{
    struct keyword_table
     *low = keyword_table, *high = LAST (keyword_table), *mid;

    struct mnemonic_table
     *low_mnem = mnemonic_table, *high_mnem = LAST (mnemonic_table), *mid_mnem;

    int c = 0;
    char lower_case[30];

    while (yytext[c])
	lower_case[c++] = tolower (yytext[c]);

    lower_case[c] = '\0';

    while (low <= high) {
	mid = low + (high - low) / 2;

	if (!(c = strcmp (mid->keyword_name, (char *) lower_case)))
	    return mid->keyword_representation;
	else if (c < 0)
	    low = mid + 1;
	else
	    high = mid - 1;
    }

    /*
     * check if       it is a mnemonic if we haven't seen one already on the
     * current line
     */
    if (!seen_mnemonic) {
	while (low_mnem <= high_mnem) {
	    mid_mnem = low_mnem + (high_mnem - low_mnem) / 2;

	    if (!(c = strcmp (mid_mnem->mnemonic_name, lower_case))) {
		last_mnem = mid_mnem;
		seen_mnemonic = TRUE;

		/*
		 * The beginning of this line should be       the beginning of the
		 * instruction
		 */
		start_instr = start_line ();

		/*
		 * The end of the mnemonic is the start       of the first
		 * operand
		 */
		start_first = save_pos ();

		return MNEMONIC;
	    }
	    else if (c < 0)
		low_mnem = mid_mnem + 1;
	    else
		high_mnem = mid_mnem - 1;
	}
    }

    return ID;
}

yywrap ()
{
    /* donno what this does */
    return 1;
}

/* all errors lead to this function */
yyerror (va_alist)
     va_dcl
{
    va_list argp;
    char *fmt;
    char message[MAX_ID_LENGTH * 2];

    va_start (argp);
    fmt = va_arg (argp, char *);
    vsprintf (message, fmt, argp);
    yywhere ();
    fprintf (stderr, "%s \n", message);

    exit (1);
}

/* where are we ? */
yywhere ()
{
    extern char input_filename[MAX_ID_LENGTH];

    if (strlen (input_filename))
	fprintf (stderr, "%s:%d: ", input_filename, yylineno);
}

/* store the current token in the buffer - char by char */
token_store ()
{
    char c;
    int i = 0;

    while ((c = yytext[i++]) != 0)
	put_buf (c);
}
