/* 
 * Please read the COPYRIGHT notice included at the top level
 */
#include <stdio.h>
#include <assert.h>

#include "doctor_t.h"
#include "lex.h"
#include "memref.h"
#include "instr.h"
#include "mnemonics_def.h"
#include "write.h"
#include "augment.h"

extern bool_t stack_ref_augment;

/* Allocates and initializes a new reference structure */
ref_t *
new_ref ()
{
    ref_t *new_reference;
    int size;

    size = sizeof (ref_t);
    new_reference = (ref_t *) malloc (size);
    memset ((void *) new_reference, 0, size);

    return new_reference;
}

/* sets up the instr->mem_ref field appropriately */
void
check_mem_ref (instr_t *instr)
{
    enum enum_ref_t explicit;

    explicit = is_explicit_mem_ref (instr);

    switch (instr->mnemonic->read_write) {
	case NO_REF:
	    /* wow, that's cool! */
	    return;
	case IMPLICIT_READ:
	    switch (instr->mnemonic->representation) {
		case CMPS:
		case CMPSB:
		case CMPSW:
		case CMPSL:
		    instr->mem_ref[0] = new_ref ();
		    instr->mem_ref[0]->implicit = TRUE;
		    instr->mem_ref[0]->read = TRUE;
		    instr->mem_ref[1] = new_ref ();
		    instr->mem_ref[1]->implicit = TRUE;
		    instr->mem_ref[1]->read = TRUE;
		    /* the second parameter doesn't really matter here */
		    get_implicit_op (instr, NULL);
		    break;
		default:
		    instr->mem_ref[0] = new_ref ();
		    instr->mem_ref[0]->implicit = TRUE;
		    instr->mem_ref[0]->read = TRUE;
		    get_implicit_op (instr, instr->mem_ref[0]);
		    break;
	    }
	    break;
	case IMPLICIT_WRITE:
	    instr->mem_ref[0] = new_ref ();
	    instr->mem_ref[0]->implicit = TRUE;
	    instr->mem_ref[0]->write = TRUE;
	    get_implicit_op (instr, instr->mem_ref[0]);
	    break;
	case MEMORY_READ:
	    if (explicit != NONE) {
		instr->mem_ref[0] = new_ref ();
		instr->mem_ref[0]->read = TRUE;

		if (explicit == FIRST)
		    instr->mem_ref[0]->source =
			get_first_op (instr);
		else if (explicit == SECOND)
		    instr->mem_ref[0]->source =
			get_second_op (instr);
		/* what about the destination ?? */
	    }
	    break;
	case MEMORY_WRITE:
	    if (explicit != NONE) {
		instr->mem_ref[0] = new_ref ();
		instr->mem_ref[0]->write = TRUE;

		if (explicit == FIRST)
		    instr->mem_ref[0]->destination =
			get_first_op (instr);
		else if (explicit == SECOND)
		    instr->mem_ref[0]->destination =
			get_second_op (instr);
		/* what about the source ?? */
	    }
	    break;
	case KNOWN:
	    switch (explicit) {
		case FIRST:
		    instr->mem_ref[0] = new_ref ();
		    instr->mem_ref[0]->read = TRUE;
		    instr->mem_ref[0]->source = get_first_op (instr);
		    break;
		case SECOND:
		    instr->mem_ref[0] = new_ref ();
		    instr->mem_ref[0]->write = TRUE;
		    instr->mem_ref[0]->destination = get_second_op (instr);
		    break;
		case NONE:
		    return;
		default:
		    assert (0);
		    break;
	    }
	    break;
	case READ_WRITE:
	    /* Got to tread carefully here */
	    switch (instr->mnemonic->representation) {
		case CALL:
		case LCALL:
		    switch (explicit) {
			case FIRST:
			    instr->mem_ref[0] = new_ref ();
			    instr->mem_ref[0]->read = TRUE;
			    instr->mem_ref[0]->source = get_first_op (instr);
			    /* break intentionally omitted */
			case NONE:
			    if (instr->mem_ref[0]) {
				instr->mem_ref[1] = new_ref ();
				instr->mem_ref[1]->write = TRUE;
				get_implicit_op (instr, instr->mem_ref[1]);
			    }
			    else {
				instr->mem_ref[0] = new_ref ();
				instr->mem_ref[0]->write = TRUE;
				get_implicit_op (instr, instr->mem_ref[0]);
			    }
			    break;
			default:
			    assert (0);
			    break;
		    }
		    break;
		case MOVS:
		case MOVSB:
		case MOVSW:
		case MOVSL:
		    instr->mem_ref[0] = new_ref ();
		    instr->mem_ref[0]->implicit = TRUE;
		    instr->mem_ref[0]->read = TRUE;
		    instr->mem_ref[1] = new_ref ();
		    instr->mem_ref[1]->implicit = TRUE;
		    instr->mem_ref[1]->write = TRUE;
		    /* The second parameter doesn't matter */
		    get_implicit_op (instr, NULL);
		    break;
		case ARPL:
		    assert (0);
		    /* the r/w is determined dynamically */
		    break;
		case POP:
		case POPL:
		case POPW:
		    switch (explicit) {
			case FIRST:
			    instr->mem_ref[0] = new_ref ();
			    instr->mem_ref[0]->write = TRUE;
			    instr->mem_ref[0]->destination =
				get_first_op (instr);
			    instr->mem_ref[1] = new_ref ();
			    instr->mem_ref[1]->read = TRUE;
			    get_implicit_op (instr, instr->mem_ref[1]);
			    break;
			case NONE:
			    instr->mem_ref[0] = new_ref ();
			    instr->mem_ref[0]->read = TRUE;
			    get_implicit_op (instr, instr->mem_ref[0]);
			    break;
			default:
			    assert (0);
			    break;
		    }
		    break;
		case PUSH:
		case PUSHW:
		case PUSHL:
		    switch (explicit) {
			case FIRST:
			    instr->mem_ref[0] = new_ref ();
			    instr->mem_ref[0]->read = TRUE;
			    instr->mem_ref[0]->source = get_first_op (instr);
			    instr->mem_ref[1] = new_ref ();
			    instr->mem_ref[1]->write = TRUE;
			    get_implicit_op (instr, instr->mem_ref[1]);
			    break;
			case NONE:
			    instr->mem_ref[0] = new_ref ();
			    instr->mem_ref[0]->write = TRUE;
			    get_implicit_op (instr, instr->mem_ref[0]);
			    break;
			default:
			    assert (0);
			    break;
		    }
		    break;
		default:
		    /* This is where most instructions fall */
		    switch (explicit) {
			case FIRST:
			    if (instr->num_ops == 1) {
				/* instructions like inc,dec etc */
				instr->mem_ref[0] = new_ref ();
				instr->mem_ref[0]->read = TRUE;
				instr->mem_ref[0]->source =
				    get_first_op (instr);
				instr->mem_ref[1] = new_ref ();
				instr->mem_ref[1]->write = TRUE;
				instr->mem_ref[1]->destination =
				    instr->mem_ref[0]->source;
			    }
			    else if (instr->num_ops == 2) {
				/* cases like mov etc */
				instr->mem_ref[0] = new_ref ();
				instr->mem_ref[0]->read = TRUE;
				instr->mem_ref[0]->source =
				    get_first_op (instr);
			    }
			    else
				assert (0);
			    break;
			case SECOND:
			    instr->mem_ref[0] = new_ref ();
			    instr->mem_ref[0]->read = TRUE;
			    instr->mem_ref[0]->source =
				get_second_op (instr);
			    instr->mem_ref[1] = new_ref ();
			    instr->mem_ref[1]->write = TRUE;
			    instr->mem_ref[1]->destination =
				get_second_op (instr);
			    break;
			case NONE:
			    break;
			default:
			    assert (0);
			    break;
		    }
		    break;
	    }
	    break;
	default:
	    assert (0);
	    break;
    }
}

/* finds out if the instruction makes an explicit memory reference */
enum enum_ref_t
is_explicit_mem_ref (instr_t *instr)
{
    enum addr_modes *mode_list;

    mode_list = instr->modes;

    if (mode_list != NULL) {
	if ((mode_list[0] == DIRECT) || (mode_list[0] == INDIRECT)
	    || ((mode_list[0] == STACKREF) && (stack_ref_augment)))
	    return FIRST;

	if ((mode_list[1] == DIRECT) || (mode_list[1] == INDIRECT)
	    || ((mode_list[0] == STACKREF) && (stack_ref_augment)))
	    return SECOND;
    }

    return NONE;
}

/*
 * Gets       the (text of the) first operand of the instruction. Also note
 * that the operand name gets overwritten, when get_first_op is called
 * again.
 */
char *
get_first_op (instr_t *instr)
{
    static char operand[MAX_OP_SIZE];
    extern char *output_buffer;
    int i, j;

    i = instr->start_first;
    j = 0;

    while (i != instr->end_first) {
	operand[j++] = output_buffer[i];

	/*
	 * An ugly hack       from getting the comments out of the second
	 * operand ..
	 */
	if ((output_buffer[i] == '/') || (output_buffer[i] == '\n')) {
	    operand[j - 1] = '\0';

	    return convert_jmp_normal (operand);
	}

	INC_OPB_PTR (i);
    }
    operand[j] = '\0';

    return convert_jmp_normal (operand);
}

/*
 * Jumps and calls have       an unusual syntax, with an asterisk preceding
 * the memory  operand. That needs to be removed to get the actual
 * operand.
 */
char *
convert_jmp_normal (char *operand)
{
    /* convert the jmp/call memory reference into the normal format */
    if (jmp_memref (operand)) {
	char *ptr;

	ptr = skip_whitespaces (operand);
	/* skip the asterisk */
	ptr++;

	return ptr;
    }
    return operand;
}

/* Gets the (text of the) second operand of the instruction */
char *
get_second_op (instr_t *instr)
{
    static char operand[MAX_OP_SIZE];
    extern char *output_buffer;
    int i, j;

    i = instr->start_second;
    j = 0;

    while (i != instr->end_second) {
	operand[j++] = output_buffer[i];

	/*
	 * An ugly hack       from getting the comments out of the second
	 * operand ..
	 */
	if ((output_buffer[i] == '/') || (output_buffer[i] == '\n')) {
	    operand[j - 1] = '\0';
	    return operand;
	}

	INC_OPB_PTR (i);
    }

    if (output_buffer[i] != '\n')
	operand[j++] = output_buffer[i];
    operand[j] = '\0';

    return operand;
}

/* The list of implicit operands */
static char *implicit_ops[] =
{
    "%es:(%di)",
    "%ds:(%si)",
    "%es:(%edi)",
    "%es:(%esi)",
    "(%sp)",
    "%d(%%sp)",
    "(%esp)",
    "%d(%%esp)",
    "%d(%%bp)",
    "%d(%%ebp)",
};

enum implicit_t {
    SOURCE_INDEX = 0, ESOURCE_INDEX, DEST_INDEX, EDEST_INDEX,
    STACK_P, OFFSET_STACK_P, ESTACK_P, EOFFSET_STACK_P,
    BASE_P, EBASE_P
};

/* implicit operands are always pointed to by the mem_op_second */
void
get_implicit_op (instr_t *instr, ref_t *ref)
{
    ref_t *ref1, *ref2;

    ref1 = instr->mem_ref[0];
    ref2 = instr->mem_ref[1];

    switch (instr->mnemonic->representation) {
	case CMPSB:
	case CMPSL:
	case CMPSW:
	    ref1->read = TRUE;
	    ref2->read = TRUE;
	    ref1->source = implicit_ops[ESOURCE_INDEX];
	    ref2->source = implicit_ops[EDEST_INDEX];
	    break;
	case MOVS:
	case MOVSB:
	case MOVSW:
	case MOVSL:
	    ref1->read = TRUE;
	    ref2->write = TRUE;
	    ref1->source = implicit_ops[ESOURCE_INDEX];
	    ref2->destination = implicit_ops[EDEST_INDEX];
	    break;
	case LEAVE:
	    instr->augment &= stack_ref_augment;
	    ref->read = TRUE;
	    ref->format_string = implicit_ops[EBASE_P];
	    ref->stack = TRUE;
	    break;
	case RET:
	case LRET:
	case IRET:
	case IRETF:
	case LIRET:
	case POP:
	case POPA:
	case POPF:
	case POPFL:
	case POPL:
	    instr->augment &= stack_ref_augment;
	    ref->read = TRUE;
	    ref->format_string = implicit_ops[EOFFSET_STACK_P];
	    ref->stack = TRUE;
	    ref->stack_offset += 4 * NUM_REGS_PUSHED;
	    break;
	case CALL:
	case LCALL:
	case ENTER:
	case INT:
	case PUSH:
	case PUSHA:
	case PUSHF:
	case PUSHAW:
	case PUSHAL:
	case PUSHFL:
	case PUSHFW:
	case PUSHL:
	case PUSHW:
	    instr->augment &= stack_ref_augment;
	    ref->format_string = implicit_ops[EOFFSET_STACK_P];
	    ref->stack = TRUE;
	    ref->write = TRUE;
	    ref->stack_offset = 4 * NUM_REGS_PUSHED -
		instr->mnemonic->operand_size;
	    break;
	case INS:
	case INSB:
	case INSW:
	case INSL:
	    ref->source = implicit_ops[EDEST_INDEX];
	    ref->write = TRUE;
	    break;
	case SCAS:
	case SCASB:
	case SCASL:
	case SCASW:
	    ref->read = TRUE;
	    ref->source = implicit_ops[EDEST_INDEX];
	    break;
	case STOS:
	case STOSB:
	case STOSW:
	case STOSL:
	    ref->write = TRUE;
	    ref->destination = implicit_ops[EDEST_INDEX];
	    break;
	case LODS:
	case LODSB:
	case LODSW:
	case LODSL:
	case OUTS:
	case OUTSB:
	case OUTSW:
	case OUTSL:
	    ref->read = TRUE;
	    ref->source = implicit_ops[ESOURCE_INDEX];
	    break;
	default:
	    assert (0);
	    break;
    }
}

/* Special instructions are very rare aren't they :-) */
void
handle_special_instr (instr_t *instr)
{
    switch (instr->mnemonic->representation) {
	default:
	    yyerror ("%s not handled yet", instr->mnemonic->mnemonic_name);
	    break;
    }
}

/*
 * checks if this is a jump memory reference - look at
 * convert_jmp_operand for more explanation.
 */
bool_t
jmp_memref (char *s)
{
    s = skip_whitespaces (s);
    return (*s == '*');
}

char *
skip_whitespaces (char *s)
{
    while ((*s == ' ') || (*s == '\t'))
	s++;

    return s;
}
